"""Interfaces for Relations product.

See also the class diagram in "doc/relationsdia2.png".

Unless explicitely stated otherwise, source and target arguments must be of
type IBrainAggregate, which comprises all information available from catalogs
on an object.
"""

from Interface import Interface, Attribute
from Products.Archetypes.interfaces.referenceengine import IReference


class IChain(Interface):
    """A non persistent object that is passed around while processing
    reference creation and deletion. Attributes 'added' and 'deleted' are lists
    of thus far created and deleted references, respectively.

    IChain is dict-like. However, an IChain creates (ordinary) dicts where
    a normal dict would raise KeyError:
    >>> chain.keys()
    []
    >>> chain['mynamespace']
    {}
    >>> chain.keys()
    ['mynamespace']
    >>> chain['mynamespace'][123] = 456
    >>> chain['mynamespace']
    {123: 456}
    """
    
    added = Attribute('added',
                      "The list of reference obects added in the current "
                      "process.")
    deleted = Attribute('deleted',
                        "The list of reference objects deleted in the current "
                        "process.")

# processor (module processor)
class IReferenceConnectionProcessor(Interface):
    """Manages the process of connecting and disconnecting references.

    The process is divided into three stages: implication, validation and
    finalization."""

    def process(context, connect=(), disconnect=()):
        """Creates and deletes references.

        Connect and disconnect arguments are triples of strings in the form
            (Source UID, Target UID, Relationship)
        where Relationship is the ID of the jig that we intend to use.

        If the process fails with an Exception, process rolls back the database
        to before this function was called.
        """

# container (module jig)
class IReferenceJig(Interface):
    """ReferenceJig is a container for implicators, validators, finalizers
    and vocabulary providers.

    It provides methods to initiate implications, do validation and
    finalization."""

    def getId():
        """Returns the jig's ID, which must be unique among jigs."""
    
    def implyOnConnect(source, target, chain):
        """Calls the PrimaryImplicator's connect method, thereby creating
        a reference between source and target.

        Appends the newly created reference to the chain and calls each
        implicator's implyOnConnect method.

        If a reference between source and target already exists for self,
        neither is a reference created, nor is the reference appended to
        chain."""

    def implyOnDisconnect(reference, chain):
        """Calls implyOnDisconnect() for each implicator.

        In case such a reference doesn't exist (anymore), this is a noop, and
        nothing is appended to chain."""


    def validateConnected(reference, chain):
        """Calls validateConnected() for each validator.

        Validator components raise exception.ValidationException if validation
        fails."""

    def validateDisconnected(reference, chain):
        """Calls validateDisconnected() for each validator.

        Validator components raise exceptions.ValidationException if validation
        fails."""


    def finalizeOnConnect(reference, chain):
        """Calls finalizeOnConnect() for each finalizer."""

    def finalizeOnDisconnect(reference, chain):
        """Calls finalizeOnDisconnect() for each finalizer."""


    def makeVocabulary(source, targets=None):
        """Calls makeVocabulary() for each vocabulary provider, forwarding
        the result of the n-th provider to the n+1-th provider, or returning
        the result if no there is no n+1-th provider.

        You may optionally provide a list of IBrainAggregates as the targets
        argument. This list is then forwarded to the first provider in a
        request to filter it."""


class IReferenceJigRegistry(Interface):
    """Holds references to ReferenceJigs, identified by their id."""

    def registerJig(jig):
        """Build a reference to jig and make it available through getJig()."""

    def getJig(id):
        """Returns the jig with the given id or raise zExceptions.NotFound if
        no such jig is known."""

    def getJigs():
        """Return the list of all registered jigs."""

    def getFolder():
        """Convenience method that returns a suitable folder for storing
        IReferenceJigs."""


#components
class IJigComponent(Interface):
    """Superinterface of all components."""

    def getJig():
        """Returns the IReferenceJig that owns this component."""

class IVocabularyProvider(IJigComponent):
    """Vocabulary providers create lists of possible targets."""

    def makeVocabulary(source, targets=None):
        """Returns a list of aggregated brains.

        The targets argument may be either None or a list of IBrainAggregates.
        None is to signal that the provider must build a list of
        IBrainAggregates itself. If targets are given, the provider must filter
        them."""

class IPrimaryImplicator(IJigComponent):
    """The primary implicator is responsible for establishing a reference in
    the ReferenceEngine.

    See jig.DefaultPrimaryImplicator for an implementation."""

    def connect(source, target):
        """If a reference between source and target does not yet exist, creates
        a new reference and returns it. Returns None if a reference already
        exists."""

    def disconnect(reference):
        """Returns None if no such reference exists. Otherwise deletes the
        reference in the db."""


class IImplicator(IJigComponent):
    """XXX"""

    def implyOnConnect(reference, chain):
        """Typically calls ReferenceJig.implyOnConnect of other jigs."""

    def implyOnDisconnect(reference, chain):
        """XXX"""


class IValidator(IJigComponent):
    """XXX"""

    def validateConnected(reference, chain):
        """Raises exceptions.ValidationException if invalid.

        Note that IValidators may not expect reference to be inside of
        chain.added. validateConnected() may be called at any time for an
        existing reference."""

    def validateDisconnected(reference, chain):
        """Raises exceptions.ValidationException if invalid.

        Note that reference represents a reference that has been deleted from
        the database. Its nonexistence is to be validated.

        The reference argument is always in chain.deleted."""


class IFinalizer(IJigComponent):
    """XXX"""

    def finalizeOnConnect(reference, chain):
        """XXX"""

    def finalizeOnDisconnect(reference, chain):
        """XXX"""

# composite hooks (module jig)
class IReferenceLayerProvider(IJigComponent):
    """Components may signal through this interface that they want to add
    behaviour to the reference's hooks. See IReferenceLayer"""

    def provideReferenceLayer(reference):
        """Return an object implementing IReferenceLayer."""


class IReferenceLayer(Interface):
    """Used to add behaviour to reference class hooks.

    Note that methods are not mandatory, they are called only when they exist.
    """

    def addHook(reference):
        """XXX"""

    def delHook(reference):
        """XXX"""

    def beforeTargetDeleteInformSource(reference):
        """XXX"""

    def beforeSourceDeleteInformTarget(reference):
        """XXX"""
    

# brain aggregates and convenience (module brain)
class IBrainAggregate(Interface):
    """Resembles a ZCatalog brain in usage, i.e. all metadata is available
    through attributes and the real object itself can be retrieved through
    getObject()."""

    def getObject():
        """Return the actual object that this aggregate holds the metadata
        for."""


class IReferenceWithBrains(IReference):
    """Minimal extension to IReference with two convenience methods added."""

    def getSourceBrain():
        """Returns an IBrainAggregate, which comprises all catalog metadata
        available for source object."""

    def getTargetBrain():
        """Returns an IBrainAggregate, which comprises all catalog metadata
        available for target object."""

