from Acquisition import ImplicitAcquisitionWrapper, aq_parent
from Globals import InitializeClass
import zExceptions

from Products import CMFCore
from Products.CMFCore.utils import getToolByName
from Products.CMFCore.ActionInformation import ActionInformation
from Products.CMFCore.Expression import Expression
from Products.Archetypes.interfaces.referenceable import IReferenceable
from Products.Archetypes.Referenceable import Referenceable
from Products.Archetypes.ReferenceEngine import Reference
from Products.Archetypes.public import *

from config import *
import interfaces
import brain
import utils

class ReferenceLayerManager(Reference):
    """Forwards hooks to ReferenceLayers, see IReferenceLayerProvider."""
    
    def addHook(self, tool, sourceObject=None, targetObject=None):
        self._forEachLayerDo('addHook', self)

    def delHook(self, tool, sourceObject=None, targetObject=None):
        self._forEachLayerDo('delHook', self)

    def beforeTargetDeleteInformSource(self):
        self._forEachLayerDo('beforeTargetDeleteInformSource', self)

    def beforeSourceDeleteInformTarget(self):
        self._forEachLayerDo('beforeSourceDeleteInformTarget', self)

    def getJig(self):
        if not getattr(self, '_v_jig', None):
            registry = getToolByName(self, REFERENCEJIGREGISTRY)
            self._v_jig = registry.getJig(self.relationship)
        return self._v_jig

    def _forEachLayerDo(self, methodName, *args, **kwargs):
        layers = self._getReferenceLayers()
        for l in layers:
            method = getattr(l, methodName, None)
            if callable(method):
                method(*args, **kwargs)

    def _getReferenceLayers(self):
        jig = self.getJig()
        providers = jig._filterByInterface(interfaces.IReferenceLayerProvider)
        return [p.provideReferenceLayer(self) for p in providers]

InitializeClass(ReferenceLayerManager)

class RLMWithBrains(ReferenceLayerManager, brain.ReferenceWithBrains):
    pass

class JigComponentBase:
    __implements__ = interfaces.IJigComponent,

    global_allow = 0 # convenience
    def getJig(self): return aq_parent(self)

class DefaultPrimaryImplicator(JigComponentBase):
    __implements__ = interfaces.IPrimaryImplicator,

    referenceClass = RLMWithBrains

    def __init__(self, jig):
        self.jig = jig
    
    def connect(self, source, target):
        ref_catalog = getToolByName(self.jig, REFERENCE_CATALOG)
        if not ref_catalog(sourceUID=source.UID, targetUID=target.UID,
                           relationship=self.jig.getId()):
            return ref_catalog.addReference(source.UID, target.UID,
                                            self.jig.getId(),
                                            self.referenceClass)

    def disconnect(self, reference):
        ref_catalog = getToolByName(self.jig, REFERENCE_CATALOG)
        if ref_catalog(sUID=reference.sourceUID, tUID=reference.targetUID,
                       relationship=self.jig.getId()):
            ref_catalog._deleteReference(reference)
            return True
        
class ReferenceJig(utils.AllowedTypesByIface, OrderedBaseFolder):
    """See IReferenceJig."""
    __implements__ = (interfaces.IReferenceJig,) +\
                     OrderedBaseFolder.__implements__

    schema = BaseSchema
    portal_type = 'Reference Jig'
    archetype_name = "Reference Rulebook"
    global_allow = 0
    content_icon = 'rulebook_icon.gif'
    allowed_interfaces = interfaces.IJigComponent, #used by AllowedTypesByIface

##     def initializeArchetype(self, **kwargs):
##         OrderedBaseFolder.initializeArchetype(self, **kwargs)

    def implyOnConnect(self, source, target, chain):
        primaryImplicator = self._getPrimaryImplicator()
        reference = primaryImplicator.connect(source, target)
        if reference is not None:
            chain.added.append(reference)
            self._forEachDo(interfaces.IImplicator, 'implyOnConnect',
                            reference, chain)

    def implyOnDisconnect(self, reference, chain):
        primaryImplicator = self._getPrimaryImplicator()
        performed = primaryImplicator.disconnect(reference)
        if performed is not None:
            chain.deleted.append(reference)
##             reference = ImplicitAcquisitionWrapper(reference, self)
            self._forEachDo(interfaces.IImplicator, 'implyOnDisconnect',
                            reference, chain)

    def validateConnected(self, reference, chain):
        self._forEachDo(interfaces.IValidator, 'validateConnected',
                        reference, chain)

    def validateDisconnected(self, reference, chain):
        self._forEachDo(interfaces.IValidator, 'validateDisconnected',
                        reference, chain)

    def finalizeOnConnect(self, reference, chain):
        self._forEachDo(interfaces.IFinalizer, 'finalizeOnConnect',
                        reference, chain)

    def finalizeOnDisconnect(self, reference, chain):
        self._forEachDo(interfaces.IFinalizer, 'finalizeOnDisconnect',
                        reference, chain)

    def makeVocabulary(self, source, targets=None):
        for obj in self._filterByInterface(interfaces.IVocabularyProvider):
            targets = obj.makeVocabulary(source, targets)
        return targets

    def _forEachDo(self, interface, methodname, *args):
        """For each contained object that is implements interface, call method
        with *args."""
        for obj in self._filterByInterface(interface):
            meth = getattr(obj, methodname)
            meth(*args)

    def _filterByInterface(self, interface):
        """Return a list of objects that implement the given interface."""
        return [obj for obj in self.objectValues()
                if interface.isImplementedBy(obj)]

    def _getPrimaryImplicator(self):
        impls = self._filterByInterface(interfaces.IPrimaryImplicator)
        if impls:
            return impls[0]
        else:
            return DefaultPrimaryImplicator(self)

registerType(ReferenceJig)


class ReferenceJigRegistry(utils.AllowedTypesByIface, BaseFolder):
    """Registry for Reference Jigs. See IReferenceJigRegistry."""
    __implements__ = (interfaces.IReferenceJigRegistry,
                      CMFCore.interfaces.portal_actions.ActionProvider) + \
                      BaseFolder.__implements__

    schema = BaseSchema
    portal_type = 'Jig Registry'
    archetype_name = "Rulebook Library"
    global_allow = 0
    allowed_interfaces = interfaces.IReferenceJig,

    ##
    # see IReferenceJigRegistry
    def registerJig(self, jig):
        self.addReference(jig, RELATIONSHIP_JIGREGISTRY)

    def getJig(self, id):
        ref_ctl = getToolByName(self, REFERENCE_CATALOG)
        uid_ctl = getToolByName(self, UID_CATALOG)

        refs = ref_ctl._queryFor(sid=self.UID(),
                                 relationship=RELATIONSHIP_JIGREGISTRY,
                                 targetId=id)
        if refs:
            return refs[0].getObject().getTargetObject()
        else:
            raise zExceptions.NotFound, "No jig with id %s" % id

    def getJigs(self):
        return self.getRefs(RELATIONSHIP_JIGREGISTRY)

    def getFolder(self):
        return self

    # Add reference to reference jigs.
    def invokeFactory(self, type_name, id, RESPONSE = None, *args, **kwargs):
        super_invokeFactory = utils.AllowedTypesByIface.invokeFactory
        v = super_invokeFactory(self, type_name, id, RESPONSE, *args, **kwargs)
        jig = getattr(self, v)
        ref_ctl = getToolByName(self, REFERENCE_CATALOG)        
        brains = ref_ctl(sourceUID=self.UID(),
                         targetUID=jig.UID(),
                         relationship=RELATIONSHIP_JIGREGISTRY)
        if len(brains) == 0: # no such ref exists
            self.addReference(jig, RELATIONSHIP_JIGREGISTRY)
        return v

    # see CMFCore.interfaces.portal_actions.ActionProvider
    def listActions(self, info=None):
        
        action = Expression(
            'string:$object_url/relations_form')
        condition = Expression(
            'python:object.relations_form_vocab(test_only=1)')
        ai = ActionInformation(id='relations',
                               title='Relations',
                               action=action,
                               condition=condition,
                               category='object_tabs',
                               permissions=('Modify portal content',),
                               visible=1)
        return ai,
        
registerType(ReferenceJigRegistry)
