import os, sys
if __name__ == '__main__':
    execfile(os.path.join(sys.path[0], 'framework.py'))

from Products.CMFTestCase import CMFTestCase
import common
common.installProducts()
CMFTestCase.setupCMFSite()

from Acquisition import aq_parent
from OFS.SimpleItem import SimpleItem
from Products.CMFCore.utils import getToolByName

from Products.Relations.config import *
import Products.Relations.interfaces as interfaces
import Products.Relations.brain as brain
import Products.Relations.jig as jigmodule
import Products.Relations.processor as processor

class TestJigRegistry(CMFTestCase.CMFTestCase):

    def afterSetUp(self):
        common.installWithinPortal(self, self.portal)
        self.registry = getToolByName(self.portal, REFERENCEJIGREGISTRY)

    def testIsAvailableAsTool(self):
        marker = []
        registry = getToolByName(self.portal, REFERENCEJIGREGISTRY, marker)
        self.assert_(registry is not marker)

    def testRegister(self):
        ttool = getToolByName(self.portal, 'portal_types')
        construct = ttool.constructContent

        # create a jig in the folder given by getFolder()
        self.loginAsPortalOwner()
        self.registry.getFolder().invokeFactory('Reference Jig', 'jig1')
        self.logout(); self.login()

        # create a jig somewhere
        construct('Reference Jig', self.folder, 'jig2')

        jig1 = getattr(self.registry.getFolder(), 'jig1')
        jig2 = getattr(self.folder, 'jig2')
        # We don't need to register jig1, because it was created inside the
        # jig registry, thus it's implicitely registered.
        self.registry.registerJig(jig2)

        for jig in jig1, jig2:
            self.assertEquals(jig, self.registry.getJig(jig.getId()))

        self.assertEquals([jig1, jig2], self.registry.getJigs())

    def testAllowedContentTypes(self):
        # only portal owner may add jigs
        self.loginAsPortalOwner()
        types = self.registry.allowedContentTypes()
        self.assertEquals(len(types), 1)
        self.assert_(str(types[0]).endswith("Reference Jig>"),
                     "%s not recognized." % str(types[0]))
        self.logout()

    def testInvokeFactory(self):
        self.loginAsPortalOwner()
        folder = self.registry.getFolder()
        folder.invokeFactory('Reference Jig', 'allowed')
        # wrong place
        self.assertRaises(ValueError,
                          self.folder.invokeFactory,
                          'Reference Jig', 'disallowed')
        # wrong type
        self.assertRaises(ValueError,
                          folder.invokeFactory,
                          'SimpleType', 'disallowed')
        self.logout()

    def testActions(self):
        # test registration as ActionProvider and the action we define

        at = getToolByName(self.portal, 'portal_actions')

        def hasaction(obj):
            category = 'object_tabs'
            if category not in at.listFilteredActions(obj).keys():
                return False
            for action in at.listFilteredActions(obj)[category]:
                if action['id'] == 'relations':
                    return action
            return False

        self.assert_(REFERENCEJIGREGISTRY in at.listActionProviders())

        # There's no action because there's no vocabularies
        self.assert_(not hasaction(self.registry))

        # We add a jig and a corresponding reference.
        self.loginAsPortalOwner()
        f = self.registry.getFolder()
        f.invokeFactory('Reference Jig', 'samerel')

        self.registry.addReference(self.registry, 'samerel')
        self.logout(); self.login()

        # Not there, jig needs to be published.
        self.assert_(not hasaction(self.registry))

        self.loginAsPortalOwner()
        self.portal.portal_workflow.doActionFor(f.samerel, 'publish')
        self.logout(); self.login()

        # Still not there, we need 'Modify portal content' on the object...
        self.assert_(not hasaction(self.registry))        

        # ... which we get now.
        self.loginAsPortalOwner()
        self.assert_(hasaction(self.registry), "'relations' action n/a")
        self.logout(); self.login()

        # What we do here is test our action's condition in cases where context
        # is not a Referenceable.
        self.assert_(not hasaction(self.folder))

        # We create a folder which is referenceable. Inside this folder, we
        # create a document which is not.
        # We want to make sure that in the context of document we don't acquire
        # the folder's vocabulary.
        self.folder.invokeFactory('SimpleFolder', 'somefolder')
        sf = self.folder.somefolder
        sf.invokeFactory('Document', 'somedocument')
        sd = sf.somedocument
        self.assertEquals(sd.UID(), sf.UID())

        sf.addReference(sf, 'samerel')
        action = hasaction(sf)
        self.assert_(action)

        self.assert_(not hasaction(sd),
                     "%r has acquired vocabulary of %r" % (sd, sf))
        

# A dummy component that implements all interfaces and stores argument values
# to its methods inside self.calls, a dict that's keyed by methodnames.
class DummyComponent(SimpleItem, jigmodule.JigComponentBase):
    __implements__ = (interfaces.IVocabularyProvider,
                      interfaces.IPrimaryImplicator,
                      interfaces.IImplicator,
                      interfaces.IValidator,
                      interfaces.IFinalizer,
                      interfaces.IReferenceLayerProvider,
                      interfaces.IReferenceLayer, # not a component interface
                      )

    # These methods will be generated. They don't do anything, just make
    # an entry in self.calls
    gen_methods =  ('makeVocabulary', 'isValidTarget',
                    'implyOnConnect', 'implyOnDisconnect',
                    'validateConnected', 'validateDisconnected',
                    'finalizeOnConnect', 'finalizeOnDisconnect',
                    'addHook', 'delHook')

    def __init__(self, jig):
        self._jig = jig
        self.calls = {}
        # We create a list for each of these methodnames, so methods
        # themselves can simply add to the list when using their name as key.
        methodnames = self.gen_methods + ('connect', 'disconnect',
                                          'provideReferenceLayer')
        for methname in methodnames:
            self.calls[methname] = []

    # We don't generate IPrimaryImplicator's 'connect' because it needs to
    # create the reference, that is, do something.
    def connect(self, source, target):
        self.calls['connect'].append((source, target))
        impl = jigmodule.DefaultPrimaryImplicator(self.getJig())
        return impl.connect(source, target)

    # Same with disconnect:
    def disconnect(self, reference):
        jig = self.getJig()
        self.calls['disconnect'].append((jig, reference))
        impl = jigmodule.DefaultPrimaryImplicator(jig)
        return impl.disconnect(reference)

    def provideReferenceLayer(self, reference):
        self.calls['provideReferenceLayer'].append((reference,))
        return self # we are our own RL

def makeMethod(methname):
    def method(self, *args):
        self.calls[methname].append(args)
    return method

# Add methods to class DummyComponent
for methname in DummyComponent.gen_methods:
    setattr(DummyComponent, methname, makeMethod(methname))

class Chain(processor.Chain): # for better readability of repr
    def __repr__(self):
        l = self.added + self.deleted
        return "<Chain with %s items: %s>" % (len(l), l)

class TestReferenceJig(CMFTestCase.CMFTestCase):

    TYPES = 'SimpleFolder', 'SimpleType'

    def afterSetUp(self):
        common.installWithinPortal(self, self.portal)

        # create some objects
        objs = common.createObjects(self, self.TYPES)
        self.brains = [brain.makeBrainAggregate(self.portal, obj.UID())
                       for obj in objs]

        # construct and register a jig
        self.jig = common.createJig(self, 'jig')

    def testForward(self):
        # Test if methods of ReferenceJig forward to component.

        # Put the dummy component inside
        component = DummyComponent(self.jig)
        self.jig._setObject('component', component)

        chain = Chain()

        # What we do now is compare the dict component.calls to what we expect
        # is called. When we compare calls['connect'] with a list containing
        # one tuple, we assert that
        #     * the method "connect" was called exactly once because the list
        #       contains one element,
        #     * the method was called with the expected arguments, namely
        #       the ones in the tuple.
        
        calls = component.calls
        self.jig.implyOnConnect(self.brains[0], self.brains[1], chain)
        self.assertEquals(calls['connect'],
                          [(self.brains[0], self.brains[1])])
        self.assertEquals(calls['implyOnConnect'],
                          [(chain.added[0], chain)])

        # provideReferenceLayer and addHook must have been called at this point
        self.assertEquals(calls['provideReferenceLayer'],
                          [(chain.added[0],)])
        self.assertEquals(calls['addHook'],
                          [(chain.added[0],)])

        # let's get the ref from the catalog and compare it to what's in added:
        ref_ctl = getToolByName(self.portal, REFERENCE_CATALOG)
        ref = ref_ctl(sUID=self.brains[0], tUID=self.brains[1],
                      relationship=self.jig.getId())[0].getObject()
        self.assertEquals(ref, chain.added[0])
        
        # basic assertions about the chain:
        self.assertEquals(len(chain.added), 1)
        self.assertNotEquals(chain.added[0], None)

        # makeVocabulary needs special care as well:
        self.jig.makeVocabulary(self.brains[0])
        self.assertEquals(calls['makeVocabulary'],
                          [(self.brains[0], None)])
        self.jig.makeVocabulary(self.brains[1], self.brains[:])
        self.assertEquals(calls['makeVocabulary'][1],
                          (self.brains[1], self.brains))

        # These methods all expect the same arguments, namely jig, ref, chain:
        for method in ('validateConnected', 'validateDisconnected',
                       'finalizeOnConnect', 'finalizeOnDisconnect',
                       'implyOnDisconnect'):
            getattr(self.jig, method)(chain.added[0], chain)
            self.assertEquals(calls[method],
                              [(chain.added[0], chain)])

        # delHook must have been called at this point
        self.assertEquals(calls['delHook'], [(chain.added[0],)])

    def testMultipleForward(self):
        # see testForward. Adds two DummyComponents.

        component1 = DummyComponent(self.jig)
        component2 = DummyComponent(self.jig)
        self.jig._setObject('component1', component1)
        self.jig._setObject('component2', component2)

        chain = Chain()
        self.jig.implyOnConnect(self.brains[0], self.brains[1], chain)

        for component in component1, component2:
            # assert that first arg is ref, second arg is chain
            self.assertEquals(component.calls['implyOnConnect'],
                              [(chain.added[0], chain)])

    def testAllowedContentTypes(self):
        types = self.jig.allowedContentTypes()
        # We don't want to make assumptions about how many components are
        # in 'components/' and registered. We just assert that there *are*
        # components that we may add.
        self.assert_(len(types) > 0)
        for ti in types:
            self.assert_(hasattr(ti, '_isTypeInformation'),
                         "%s not a type information." % ti)

    def testInvokeFactory(self):
        ti = self.jig.allowedContentTypes()[0]
        self.jig.invokeFactory(ti.id, 'allowed')
        # wrong place
        self.assertRaises(ValueError,
                          self.folder.invokeFactory,
                          ti.id, 'disallowed')
        # wrong type
        self.assertRaises(ValueError,
                          self.jig.invokeFactory,
                          self.TYPES[0], 'disallowed')

    def testRenameJigInRegistry(self):
        registry = getToolByName(self.portal, REFERENCEJIGREGISTRY)
        self.loginAsPortalOwner()
        registry.invokeFactory('Reference Jig', 'some_id')

        # ensure object has a _p_jar
        get_transaction().commit(1)

        registry.manage_renameObject('some_id', 'something_else')

    def testDefaultPrimaryImplicator(self):
        ref_ctl = getToolByName(self.portal, REFERENCE_CATALOG)

        impl = jigmodule.DefaultPrimaryImplicator(self.jig)
        
        # connect 0 and 1
        ref = impl.connect(self.brains[0], self.brains[1])
        self.assert_(ref is not None)
        brains = ref_ctl(sUID=self.brains[0].UID, tUID=self.brains[1].UID,
                         relationship=self.jig.getId())
        self.assertEquals(len(brains), 1)
        self.assertEquals(ref, brains[0].getObject())

        # now do the same again, connect must return None now
        no_ref = impl.connect(self.brains[0], self.brains[1])
        self.assertEquals(no_ref, None)

        # let's delete it now
        self.assert_(impl.disconnect(ref) is not None)
        brains = ref_ctl(sUID=self.brains[0].UID, tUID=self.brains[1].UID,
                         relationship=self.jig.getId())
        self.assertEquals(len(brains), 0)

        # let's delete it again, disconnect returns None
        self.assert_(impl.disconnect(ref) is None)

    def testReferenceWithBrains(self):
        # tests class ReferenceWithBrains of module brain (XXX)
        chain = Chain()
        self.jig.implyOnConnect(self.brains[0], self.brains[1], chain)
        ref = self.brains[0].getObject().getReferenceImpl()[0]
        self.assertEquals(ref.getSourceBrain(), self.brains[0])
        self.assertEquals(ref.getTargetBrain(), self.brains[1])
        self.assertEquals(ref.getSourceObject(), self.brains[0].getObject())
        self.assertEquals(ref.getTargetObject(), self.brains[1].getObject())


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    tests = TestJigRegistry, TestReferenceJig
    for test in tests:
        suite.addTest(makeSuite(test))
    return suite

if __name__ == '__main__':
    framework()
